"""
MIT License

Copyright (c) 2023 Semantic Partners Ltd

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import requests
from pyanzo import AnzoClient
from rdflib import Graph, ConjunctiveGraph, Literal, URIRef
from requests import ConnectTimeout, Response, HTTPError, RequestException, ConnectionError
from bs4 import BeautifulSoup
import logging
from namespace import MUST


# https://github.com/Semantic-partners/mustrd/issues/73
def manage_anzo_response(response: Response) -> str:
    content_string = response.content.decode("utf-8")
    if response.status_code == 200:
        return content_string
    elif response.status_code == 403:
        html = BeautifulSoup(content_string, 'html.parser')
        title_tag = html.title.string
        raise HTTPError(f"Anzo authentication error, status code: {response.status_code}, content: {title_tag}")
    else:
        raise RequestException(f"Anzo error, status code: {response.status_code}, content: {content_string}")


def query_with_bindings(bindings: dict, when: str) -> str:
    values = ""
    for key, value in bindings.items():
        values += f"VALUES ?{key} {{{value.n3()}}} "
    split_query = when.lower().split("where {", 1)
    return f"{split_query[0].strip()} WHERE {{ {values} {split_query[1].strip()}"

def execute_select (triple_store: dict,  when: str, bindings: dict = None) -> str:
    try:
        if bindings:
            when = query_with_bindings(bindings, when)
        when = when.replace("${fromSources}", f"FROM <{triple_store['input_graph']}>\nFROM <{triple_store['output_graph']}>").replace(
        "${targetGraph}", f"<{triple_store['output_graph']}>")
        data = {'datasourceURI': triple_store['gqe_uri'], 'query': when,
                 'skipCache': 'true'}
        url = f"https://{triple_store['url']}:{triple_store['port']}/sparql?format=application/sparql-results+json"
        return manage_anzo_response(requests.post(url=url,
                                                  auth=(triple_store['username'], triple_store['password']),
                                                  data=data,
                                                  verify=False))
    except (ConnectionError, TimeoutError, HTTPError, ConnectTimeout):
        raise

def execute_update(triple_store: dict, when: str, bindings: dict = None) -> Graph:
    logging.debug(f"updating in anzo! {triple_store=} {when=}")
    input_graph = triple_store['input_graph']
    output_graph = triple_store['output_graph']

    substituted_query = when.replace("${usingSources}", f"USING <{triple_store['input_graph']}> \nUSING <{triple_store['output_graph']}>").replace(
        "${targetGraph}", f"<{output_graph}>")
   
    data = {'datasourceURI': triple_store['gqe_uri'], 'update': substituted_query,
                'default-graph-uri': input_graph, 'skipCache': 'true'}
    url = f"https://{triple_store['url']}:{triple_store['port']}/sparql?format=ttl"
    response = manage_anzo_response(requests.post(url=url,
            auth=(triple_store['username'],
                triple_store['password']),
            data=data,
            verify=False))
    logging.debug(f'response {response}')
    check_data = {'datasourceURI': triple_store['gqe_uri'], 'query': "construct {?s ?p ?o} { ?s ?p ?o }",
                'default-graph-uri': output_graph, 'skipCache': 'true'}
    everything_response = manage_anzo_response(requests.post(url=url,
            auth=(triple_store['username'],
                triple_store['password']),
            data=check_data,
            verify=False))
    # todo deal with error responses
    new_graph = Graph().parse(data=everything_response)
    logging.debug(f"new_graph={new_graph.serialize(format='ttl')}")
    return new_graph


def execute_construct(triple_store: dict, when: str, bindings: dict = None) -> Graph:
    try:
        if bindings:
            when = query_with_bindings(bindings, when)
        data = {'datasourceURI': triple_store['gqe_uri'], 'query': when,
                'default-graph-uri': triple_store['input_graph'], 'skipCache': 'true'}
        url = f"https://{triple_store['url']}:{triple_store['port']}/sparql?format=ttl"
        response = requests.post(url=url,
            auth=(triple_store['username'],
                triple_store['password']),
            data=data,
            verify=False)
        logging.debug(f'response {response}')
        return Graph().parse(data=manage_anzo_response(response))
    except (ConnectionError, TimeoutError, HTTPError, ConnectTimeout) as e:
        logging.error(f'response {e}')
        raise


# Get Given or then from the content of a graphmart
def get_spec_component_from_graphmart(triple_store: dict, graphmart: URIRef, layer: URIRef = None) -> ConjunctiveGraph:
    try:
        anzo_client = AnzoClient(triple_store['url'], triple_store['port'], 
                                 username=triple_store['username'],
                                 password=triple_store['password'])
        return anzo_client.query_graphmart(graphmart=graphmart,
                                           data_layers=layer,
                                           query_string="CONSTRUCT {?s ?p ?o} WHERE {?s ?p ?o}",
                                           skip_cache=True).as_quad_store().as_rdflib_graph()
    except RuntimeError as e:
        raise ConnectionError(f"Anzo connection error, {e}")


def get_query_from_querybuilder(triple_store: dict, folder_name: Literal, query_name: Literal) -> str:
    query = f"""SELECT ?query WHERE {{
        graph ?queryFolder {{
            ?bookmark a <http://www.cambridgesemantics.com/ontologies/QueryPlayground#QueryBookmark>;
                        <http://openanzo.org/ontologies/2008/07/System#query> ?query;
                        <http://purl.org/dc/elements/1.1/title> "{query_name}"
            }}
            ?queryFolder a <http://www.cambridgesemantics.com/ontologies/QueryPlayground#QueryFolder>;
                        <http://purl.org/dc/elements/1.1/title> "{folder_name}"
    }}"""
    anzo_client = AnzoClient(triple_store['url'], triple_store['port'], 
                             username=triple_store['username'],
                             password=triple_store['password'])
    
    result = anzo_client.query_journal(query_string=query).as_table_results().as_record_dictionaries()
    if len(result) == 0:
        raise FileNotFoundError(f"Query {query_name} not found in folder {folder_name}")
    return result[0].get("query")


# https://github.com/Semantic-partners/mustrd/issues/102
def get_query_from_step(triple_store: dict, query_step_uri: URIRef) -> str:
    query = f"""SELECT ?stepUri ?query WHERE {{
        BIND(<{query_step_uri}> as ?stepUri)
            ?stepUri a <http://cambridgesemantics.com/ontologies/Graphmarts#Step>;
                     <http://cambridgesemantics.com/ontologies/Graphmarts#transformQuery> ?query
    }}
    # """
    anzo_client = AnzoClient(triple_store['url'], triple_store['port'], 
                             username=triple_store['username'],
                             password=triple_store['password'])
    record_dictionaries = anzo_client.query_journal(query_string=query).as_table_results().as_record_dictionaries()

    return record_dictionaries[0].get(
        "query")

def get_queries_from_templated_step(triple_store: dict, query_step_uri: URIRef) -> dict:

    query = f"""SELECT ?stepUri ?param_query ?query_template WHERE {{
        BIND(<{query_step_uri}> as ?stepUri)
            ?stepUri    a <http://cambridgesemantics.com/ontologies/Graphmarts#Step> ;
   					    <http://cambridgesemantics.com/ontologies/Graphmarts#parametersTemplate> ?param_query ;
					    <http://cambridgesemantics.com/ontologies/Graphmarts#template> ?query_template .
    }}
    """
    anzo_client = AnzoClient(triple_store['url'], triple_store['port'], 
                             username=triple_store['username'],
                             password=triple_store['password'])
    record_dictionaries = anzo_client.query_journal(query_string=query).as_table_results().as_record_dictionaries()
    return record_dictionaries[0]


def get_queries_for_layer(triple_store: dict, graphmart_layer_uri: URIRef):
    query = f"""PREFIX graphmarts: <http://cambridgesemantics.com/ontologies/Graphmarts#>
    PREFIX anzo: <http://openanzo.org/ontologies/2008/07/Anzo#>
SELECT ?query ?param_query ?query_template
  {{ <{graphmart_layer_uri}> graphmarts:step ?step .
  ?step         anzo:index ?index ;
                anzo:orderedValue ?query_step .
  ?query_step graphmarts:enabled true ; 
  OPTIONAL {{  ?query_step
   				graphmarts:parametersTemplate ?param_query ;
           		graphmarts:template ?query_template ;
      . }}
  OPTIONAL {{  ?query_step
   				graphmarts:transformQuery ?query ;
      . }}
  }}
  ORDER BY ?index"""
    anzo_client = AnzoClient(triple_store['url'], triple_store['port'], 
                             username=triple_store['username'],
                             password=triple_store['password'])
    return anzo_client.query_journal(query_string=query).as_table_results().as_record_dictionaries()


def upload_given(triple_store: dict, given: Graph):
    logging.debug(f"upload_given {triple_store} {given}")

    try:
        input_graph = triple_store['input_graph']
        output_graph = triple_store['output_graph']
        clear_graph(triple_store, input_graph)
        clear_graph(triple_store, output_graph)
        serialized_given = given.serialize(format="nt")
        insert_query = f"INSERT DATA {{graph <{triple_store['input_graph']}>{{{serialized_given}}}}}"
        data = {'datasourceURI': triple_store['gqe_uri'], 'update': insert_query}
        response = requests.post(url=f"https://{triple_store['url']}:{triple_store['port']}/sparql",
                                    auth=(triple_store['username'], triple_store['password']), data=data, verify=False)
        manage_anzo_response(response)
    except (ConnectionError, TimeoutError, HTTPError, ConnectTimeout):
        raise


def clear_graph(triple_store: dict, graph_uri: str):
    try:
        clear_query = f"CLEAR GRAPH <{graph_uri}>"
        data = {'datasourceURI': triple_store['gqe_uri'], 'update': clear_query}
        url = f"https://{triple_store['url']}:{triple_store['port']}/sparql"
        response = requests.post(url=url,
                                 auth=(triple_store['username'], triple_store['password']), data=data, verify=False)
        manage_anzo_response(response)
    except (ConnectionError, TimeoutError, HTTPError, ConnectTimeout):
        raise




"""
MIT License

Copyright (c) 2023 Semantic Partners Ltd

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from typing import Union
from itertools import groupby
from jinja2 import Environment, FileSystemLoader
from dataclasses import dataclass
from enum import Enum


class testType(Enum):
    MUSTRD = "Mustrd"
    PYTEST = "Pytest"


class testStatus(Enum):
    PASSED = "passed"
    FAILED = "failed"
    SKIPPED = "skipped"


TEMPLATE_FOLDER = "src/templates/"


RESULT_LIST_MD_TEMPLATE = "md_ResultList_template.jinja"
RESULT_LIST_LEAF_MD_TEMPLATE = "md_ResultList_leaf_template.jinja"


@dataclass
class TestResult:
    test_name: str
    class_name: str
    module_name: str
    status: str
    is_mustrd: bool
    type: str

    def __init__(self, test_name: str, class_name: str, module_name: str, status: str, is_mustrd: bool):
        self.test_name = test_name
        self.class_name = class_name
        self.module_name = module_name
        self.status = status
        self.is_mustrd = is_mustrd
        self.type = testType.MUSTRD.value if self.is_mustrd else testType.PYTEST.value


@dataclass
class Stats:
    count: int
    success_count: int
    fail_count: int
    skipped_count: int

    def __init__(self, count: int, success_count: int, fail_count: int, skipped_count: int):
        self.count = count
        self.success_count = success_count
        self.fail_count = fail_count
        self.skipped_count = skipped_count


def get_result_list(test_results: list[TestResult], *group_functions):
    if len(group_functions) > 0:
        return list(map(lambda key_group:
                        ResultList(key_group[0],
                                   get_result_list(key_group[1],
                                   *group_functions[1:]),
                                   len(group_functions) == 1),
                        groupby(sorted(test_results, key=group_functions[0]), group_functions[0])))
    else:
        return list(test_results)


@dataclass
class ResultList:
    name: str
    stats: Stats
    is_leaf: bool
    # string type to workaround recursive call
    result_list: Union[list['ResultList'], list[TestResult]]

    def __init__(self, name: str, result_list: Union[list['ResultList'], list[TestResult]], is_leaf: bool = False):
        self.name = name
        self.result_list = result_list
        self.is_leaf = is_leaf
        self.compute_stats()

    def compute_stats(self):
        count, success_count, fail_count, skipped_count = 0, 0, 0, 0

        if self.is_leaf:
            count = len(self.result_list)
            success_count = len(
                list(filter(lambda x: x.status == testStatus.PASSED.value, self.result_list)))
            fail_count = len(
                list(filter(lambda x: x.status == testStatus.FAILED.value, self.result_list)))
            skipped_count = len(
                list(filter(lambda x: x.status == testStatus.SKIPPED.value, self.result_list)))
        else:
            for test_results in self.result_list:
                sub_count, sub_success_count, sub_fail_count, sub_skipped_count = test_results.compute_stats()
                count += sub_count
                success_count += sub_success_count
                fail_count += sub_fail_count
                skipped_count += sub_skipped_count

        self.stats = Stats(count, success_count, fail_count, skipped_count)
        return count, success_count, fail_count, skipped_count

    def render(self):
        environment = Environment(loader=FileSystemLoader(TEMPLATE_FOLDER))
        template = RESULT_LIST_LEAF_MD_TEMPLATE if self.is_leaf else RESULT_LIST_MD_TEMPLATE
        return environment.get_template(template).render(result_list=self.result_list, environment=environment)



"""
MIT License

Copyright (c) 2023 Semantic Partners Ltd

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import os
from typing import Tuple, List

import tomli
from rdflib.plugins.parsers.notation3 import BadSyntax

import logger_setup
from dataclasses import dataclass

from pyparsing import ParseException
from pathlib import Path
from requests import ConnectionError, ConnectTimeout, HTTPError, RequestException

from rdflib import Graph, URIRef, RDF, XSD, SH, Literal

from rdflib.compare import isomorphic, graph_diff
import pandas

from namespace import MUST
import requests
import json
from pandas import DataFrame

from spec_component import TableThenSpec, parse_spec_component, WhenSpec, ThenSpec
from utils import  is_json
from colorama import Fore, Style
from tabulate import tabulate
from collections import defaultdict
from pyshacl import validate
import logging 
from http.client import HTTPConnection
from steprunner import upload_given, run_when

log = logger_setup.setup_logger(__name__)

requests.packages.urllib3.disable_warnings()
requests.packages.urllib3.util.ssl_.DEFAULT_CIPHERS += ':HIGH:!DH:!aNULL'

logging.basicConfig(format='%(asctime)s %(message)s', level=logging.INFO)


def debug_requests_on():
    '''Switches on logging of the requests module.'''
    HTTPConnection.debuglevel = 1

    logging.basicConfig()
    logging.getLogger().setLevel(logging.DEBUG)
    requests_log = logging.getLogger("requests.packages.urllib3")
    requests_log.setLevel(logging.DEBUG)
    requests_log.propagate = True

def debug_requests_off():
    '''Switches off logging of the requests module, might be some side-effects'''
    HTTPConnection.debuglevel = 0

    root_logger = logging.getLogger()
    root_logger.setLevel(logging.WARNING)
    root_logger.handlers = []
    requests_log = logging.getLogger("requests.packages.urllib3")
    requests_log.setLevel(logging.WARNING)
    requests_log.propagate = False

debug_requests_off()

@dataclass
class Specification:
    spec_uri: URIRef
    triple_store: dict
    given: Graph
    when: WhenSpec
    then: ThenSpec


@dataclass
class GraphComparison:
    in_expected_not_in_actual: Graph
    in_actual_not_in_expected: Graph
    in_both: Graph


@dataclass
class SpecResult:
    spec_uri: URIRef
    triple_store: URIRef


@dataclass
class SpecPassed(SpecResult):
    pass


@dataclass()
class SpecPassedWithWarning(SpecResult):
    warning: str


@dataclass
class SelectSpecFailure(SpecResult):
    table_comparison: pandas.DataFrame
    message: str


@dataclass
class ConstructSpecFailure(SpecResult):
    graph_comparison: GraphComparison


@dataclass
class UpdateSpecFailure(SpecResult):
    graph_comparison: GraphComparison


@dataclass
class SparqlParseFailure(SpecResult):
    exception: ParseException


@dataclass
class SparqlExecutionError(SpecResult):
    exception: Exception


@dataclass
class TripleStoreConnectionError(SpecResult):
    exception: ConnectionError


@dataclass
class SpecSkipped(SpecResult):
    message: str


@dataclass
class SparqlAction:
    query: str


@dataclass
class SelectSparqlQuery(SparqlAction):
    pass


@dataclass
class ConstructSparqlQuery(SparqlAction):
    pass


@dataclass
class UpdateSparqlQuery(SparqlAction):
    pass


# https://github.com/Semantic-partners/mustrd/issues/19

def validate_specs(run_config: dict, triple_stores: List, shacl_graph: Graph, ont_graph: Graph)\
        -> Tuple[List, Graph, List]:
    spec_graph = Graph()
    subject_uris = set()
    focus_uris = set()
    invalid_specs = []
  #  ttl_files = list(spec_path.glob('**/*.mustrd.ttl'))
    ttl_files = list(run_config['spec_path'].glob('**/*.mustrd.ttl'))
    ttl_files.sort()
    log.info(f"Found {len(ttl_files)} mustrd.ttl files in {run_config['spec_path']}")

    for file in ttl_files:
        error_messages = []

        log.info(f"Parse: {file}")
        try:
            file_graph = Graph().parse(file)
        except BadSyntax as e:
            template = "An exception of type {0} occurred when trying to parse a spec file. Arguments:\n{1!r}"
            message = template.format(type(e).__name__, e.args)
            log.error(message)
            error_messages += [f"Could not extract spec from {file} due to exception of type "
                               f"{type(e).__name__} when parsing file"]
            continue
        # run shacl validation
        conforms, results_graph, results_text = validate(file_graph,
                                                         shacl_graph=shacl_graph,
                                                         ont_graph=ont_graph,
                                                         inference='none',
                                                         abort_on_first=False,
                                                         allow_infos=False,
                                                         allow_warnings=False,
                                                         meta_shacl=False,
                                                         advanced=True,
                                                         js=False,
                                                         debug=False)
        if not conforms:
            for msg in results_graph.objects(predicate=SH.resultMessage):
                log.warning(f"{file_graph}")
                log.warning(f"{msg} File: {file.name}")
                error_messages += [f"{msg} File: {file.name}"]

        # collect a list of uris of the tests in focus
        for focus_uri in file_graph.subjects(predicate=MUST.focus, object=Literal("true", datatype=XSD.boolean)):
            if focus_uri in focus_uris:
                focus_uri = URIRef(str(focus_uri) + "_DUPLICATE")
            focus_uris.add(focus_uri)

        # make sure there are no duplicate test IRIs in the files
        for subject_uri in file_graph.subjects(RDF.type, MUST.TestSpec):
            if subject_uri in subject_uris:
                log.warning(f"Duplicate subject URI found: {file.name} {subject_uri}. File will not be parsed.")
                error_messages += [f"Duplicate subject URI found in {file.name}."]
                subject_uri = URIRef(str(subject_uri) + "_DUPLICATE")
            if len(error_messages) > 0:
                error_messages.sort()
                error_message = "\n".join(msg for msg in error_messages)
                invalid_specs += [SpecSkipped(subject_uri, triple_store["type"], error_message) for triple_store in
                                triple_stores]
            else:
                subject_uris.add(subject_uri)
                this_spec_graph = Graph()
                this_spec_graph.parse(file)
                spec_uris_in_this_file = list(this_spec_graph.subjects(RDF.type, MUST.TestSpec))
                for spec in spec_uris_in_this_file:
                    tripleToAdd = [spec, MUST.specSourceFile, Literal(file)]
                    # print(f"adding {tripleToAdd}")
                    this_spec_graph.add(tripleToAdd)
                # print(f"beforeadd: {spec_graph}" )
                # print(f"beforeadd: {str(this_spec_graph.serialize())}" )
                spec_graph += this_spec_graph
                # print(f"afteradd: {str(spec_graph.serialize())}" )


    sourceFiles = list(spec_graph.subject_objects(MUST.specSourceFile))
    # print(f"sourceFiles: {sourceFiles}")

    valid_spec_uris = list(spec_graph.subjects(RDF.type, MUST.TestSpec))

    if focus_uris:
        invalid_focus_specs = []
        for spec in invalid_specs:
            if spec.spec_uri in focus_uris:
                invalid_focus_specs += [spec]
                focus_uris.remove(spec.spec_uri)
        log.info(f"Collected {len(focus_uris)} focus test spec(s)")
        return focus_uris, spec_graph, invalid_focus_specs
    else:
        log.info(f"Collected {len(valid_spec_uris)} valid test spec(s)")   
        return valid_spec_uris, spec_graph, invalid_specs


def get_specs(spec_uris: List[URIRef], spec_graph: Graph, triple_stores: List[dict],
              run_config: dict):
    specs = []
    skipped_results = []
    try:
        for triple_store in triple_stores:
            if "error" in triple_store:
                log.error(f"{triple_store['error']}. No specs run for this triple store.")
                skipped_results += [SpecSkipped(spec_uri, triple_store['type'], triple_store['error']) for spec_uri in
                                    spec_uris]
            else:
                for spec_uri in spec_uris:
                    try:
                        specs += [get_spec(spec_uri, spec_graph, run_config, triple_store)]
                    except (ValueError, FileNotFoundError, ConnectionError) as e:
                        skipped_results += [SpecSkipped(spec_uri, triple_store['type'], e)]

    except (BadSyntax, FileNotFoundError) as e:
        template = "An exception of type {0} occurred when trying to parse the triple store configuration file. " \
                   "Arguments:\n{1!r}"
        message = template.format(type(e).__name__, e.args)
        log.error(message)
        log.error("No specifications will be run.")

    log.info(f"Extracted {len(specs)} specifications that will be run")
    return specs, skipped_results


def run_specs(specs) -> List[SpecResult]:
    results = []
    # https://github.com/Semantic-partners/mustrd/issues/115
    for specification in specs:
        results.append(run_spec(specification))
    return results


def get_spec(spec_uri: URIRef, spec_graph: Graph, run_config: dict, mustrd_triple_store: dict = None) -> Specification:
    try:
        if mustrd_triple_store is None:
            mustrd_triple_store = {"type": MUST.RdfLib}
        components = []
        for predicate in MUST.given, MUST.when, MUST.then:
            components.append(parse_spec_component(subject=spec_uri,
                                                predicate=predicate,
                                                spec_graph=spec_graph,
                                                run_config=run_config,
                                                mustrd_triple_store=mustrd_triple_store))


        # https://github.com/Semantic-partners/mustrd/issues/92
        return Specification(spec_uri, mustrd_triple_store, components[0].value, components[1], components[2])
    
    except (ValueError, FileNotFoundError) as e:
        template = "An exception of type {0} occurred. Arguments:\n{1!r}"
        message = template.format(type(e).__name__, e.args)
        log.exception(message)
        raise
    except ConnectionError as e:
        log.error(e)
        raise


def check_result(spec, result):
    if type(spec.then) == TableThenSpec:
        return table_comparison(result, spec)
    else:
        graph_compare = graph_comparison(spec.then.value, result)
        if isomorphic(result, spec.then.value):
            return SpecPassed(spec.spec_uri, spec.triple_store["type"])
        else:
            if spec.when[0].queryType == MUST.ConstructSparql:
                return ConstructSpecFailure(spec.spec_uri, spec.triple_store["type"], graph_compare)
            else:
                return UpdateSpecFailure(spec.spec_uri, spec.triple_store["type"], graph_compare)


def run_spec(spec: Specification) -> SpecResult:
    spec_uri = spec.spec_uri
    triple_store = spec.triple_store
    # close_connection = True
    log.debug(f"run_when {spec_uri=}, {triple_store=}, {spec.given=}, {spec.when=}, {spec.then=}")
    if spec.given:
        given_as_turtle = spec.given.serialize(format="turtle")
        log.debug(f"{given_as_turtle}")
        upload_given(triple_store, spec.given)
    else:
        if triple_store['type'] == MUST.RdfLib:
            return SpecSkipped(spec_uri, triple_store['type'], "Unable to run Inherited State tests on Rdflib")
    try:
        for when in spec.when:
            log.info(f"Running {when.queryType} spec {spec_uri} on {triple_store['type']}")
            try:
                result = run_when(spec_uri, triple_store, when)
            except ParseException as e:
                return SparqlParseFailure(spec_uri, triple_store["type"], e)
            except NotImplementedError as ex:
                return SpecSkipped(spec_uri, triple_store["type"], ex.args[0])
        return check_result(spec, result)
    except (ConnectionError, TimeoutError, HTTPError, ConnectTimeout, OSError) as e:
        # close_connection = False
        template = "An exception of type {0} occurred. Arguments:\n{1!r}"
        message = template.format(type(e).__name__, e.args)
        log.error(message)
        return TripleStoreConnectionError(spec_uri, triple_store["type"], message)
    except (TypeError, RequestException) as e:
        log.error(f"{type(e)} {e}")
        return SparqlExecutionError(spec_uri, triple_store["type"], e)

    # https://github.com/Semantic-partners/mustrd/issues/78
    # finally:
    #     if type(mustrd_triple_store) == MustrdAnzo and close_connection:
    #         mustrd_triple_store.clear_graph()



def get_triple_store_graph(triple_store_graph_path: Path):
    secret_path = triple_store_graph_path.parent / Path(triple_store_graph_path.stem + "_secrets" + triple_store_graph_path.suffix)
    return Graph().parse(triple_store_graph_path).parse(secret_path)

def get_triple_stores(triple_store_graph: Graph) -> list[dict]:
    triple_stores = []
    for triple_store_config, rdf_type, triple_store_type in triple_store_graph.triples((None, RDF.type, None)):
        triple_store = {}
        # Local rdf lib triple store
        if triple_store_type == MUST.RdfLibConfig:
            triple_store["type"] = MUST.RdfLib
        # Anzo graph via anzo
        elif triple_store_type == MUST.AnzoConfig:
            triple_store["type"] = MUST.Anzo
            triple_store["url"] = triple_store_graph.value(subject=triple_store_config, predicate=MUST.url)
            triple_store["port"] = triple_store_graph.value(subject=triple_store_config, predicate=MUST.port)
            try:
                triple_store["username"] = str(triple_store_graph.value(subject=triple_store_config, predicate=MUST.username))
                triple_store["password"] = str(triple_store_graph.value(subject=triple_store_config, predicate=MUST.password))
            except (FileNotFoundError, ValueError) as e:
                triple_store["error"] = e
            triple_store["gqe_uri"] = triple_store_graph.value(subject=triple_store_config, predicate=MUST.gqeURI)
            triple_store["input_graph"] = triple_store_graph.value(subject=triple_store_config,
                                                                   predicate=MUST.inputGraph)
            triple_store["output_graph"] = triple_store_graph.value(subject=triple_store_config,
                                                                   predicate=MUST.outputGraph)
            try:
                check_triple_store_params(triple_store, ["url", "port", "username", "password", "input_graph"])
            except ValueError as e:
                triple_store["error"] = e
        # GraphDB
        elif triple_store_type == MUST.GraphDbConfig:
            triple_store["type"] = MUST.GraphDb
            triple_store["url"] = triple_store_graph.value(subject=triple_store_config, predicate=MUST.url)
            triple_store["port"] = triple_store_graph.value(subject=triple_store_config, predicate=MUST.port)
            try:
                triple_store["username"] = str(triple_store_graph.value(subject=triple_store_config, predicate=MUST.username))
                triple_store["password"] = str(triple_store_graph.value(subject=triple_store_config, predicate=MUST.password))
            except (FileNotFoundError, ValueError) as e:
                log.error(f"Credential retrieval failed {e}")
                triple_store["error"] = e
            triple_store["repository"] = triple_store_graph.value(subject=triple_store_config,
                                                                  predicate=MUST.repository)
            triple_store["input_graph"] = triple_store_graph.value(subject=triple_store_config,
                                                                   predicate=MUST.inputGraph)

            try:
                check_triple_store_params(triple_store, ["url", "port", "repository"])
            except ValueError as e:
                triple_store["error"] = e
        else:
            triple_store["type"] = triple_store_type
            triple_store["error"] = f"Triple store not implemented: {triple_store_type}"

        triple_stores.append(triple_store)
    return triple_stores


def check_triple_store_params(triple_store: dict, required_params: List[str]):
    missing_params = [param for param in required_params if triple_store.get(param) is None]
    if missing_params:
        raise ValueError(f"Cannot establish connection to {triple_store['type']}. "
                         f"Missing required parameter(s): {', '.join(missing_params)}.")


def get_credential_from_file(triple_store_name: URIRef, credential: str, config_path: Literal) -> str:
    log.info(f"get_credential_from_file {triple_store_name}, {credential}, {config_path}")
    if config_path is None:
        raise ValueError(f"Cannot establish connection defined in {triple_store_name}. "
                         f"Missing required parameter: {credential}.")
    # if os.path.isrelative(config_path)
    # project_root = get_project_root()
    path = Path(config_path)
    log.info(f"get_credential_from_file {path}")

    if not os.path.isfile(path):
        log.error(f"couldn't find {path}")
        raise FileNotFoundError(f"Credentials config file not found: {path}")
    try:
        with open(path, "rb") as f:
            config = tomli.load(f)
    except tomli.TOMLDecodeError as e:
        log.error(f"config error {path} {e}")
        raise ValueError(f"Error reading credentials config file: {e}")
    return config[str(triple_store_name)][credential]

# Convert sparql json query results as defined in https://www.w3.org/TR/rdf-sparql-json-res/
def json_results_to_panda_dataframe(result: str) -> pandas.DataFrame:
    json_result = json.loads(result)
    frames = DataFrame()
    for binding in json_result["results"]["bindings"]:
        columns = []
        values = []
        for key in binding:
            value_object = binding[key]
            columns.append(key)
            values.append(str(value_object["value"]))
            columns.append(key + "_datatype")
            if "type" in value_object and value_object["type"] == "literal":
                literal_type = str(XSD.string)
                if "datatype" in value_object:
                    literal_type = value_object["datatype"]
                values.append(literal_type)
            else:
                values.append(str(XSD.anyURI))

        frames = pandas.concat(objs=[frames, pandas.DataFrame([values], columns=columns)], ignore_index=True)
        frames.fillna('', inplace=True)

        if frames.size == 0:
            frames = pandas.DataFrame()
    return frames


# https://github.com/Semantic-partners/mustrd/issues/110
# https://github.com/Semantic-partners/mustrd/issues/52
def table_comparison(result: str, spec: Specification) -> SpecResult:
    warning = None
    order_list = ["order by ?", "order by desc", "order by asc"]
    ordered_result = any(pattern in spec.when[0].value.lower() for pattern in order_list)
    then = spec.then.value
    try:
        if is_json(result):
            df = json_results_to_panda_dataframe(result)
            columns = list(df.columns)
        else:
            raise ParseException
        sorted_columns = sorted(columns)
        sorted_then_cols = sorted(list(then))
        if not df.empty:

            if not ordered_result:
                df.sort_values(by=columns[::2], inplace=True)
                df.reset_index(inplace=True, drop=True)
                if spec.then.ordered:
                    warning = f"sh:order in {spec.spec_uri} is ignored, no ORDER BY in query"
                    log.warning(warning)

            # Scenario 1: expected no result but got a result
            if then.empty:
                message = f"Expected 0 row(s) and 0 column(s), got {df.shape[0]} row(s) and {round(df.shape[1] / 2)} column(s)"
                empty_then = create_empty_dataframe_with_columns(df)
                df_diff = empty_then.compare(df, result_names=("expected", "actual"))

            else:
                # Scenario 2: expected a result and got a result
                # pandas.set_option('display.max_columns', None)
                message = f"Expected {then.shape[0]} row(s) and {round(then.shape[1] / 2)} column(s), " \
                          f"got {df.shape[0]} row(s) and {round(df.shape[1] / 2)} column(s)"
                if ordered_result is True and not spec.then.ordered:
                    message += ". Actual result is ordered, must:then must contain sh:order on every row."
                    return SelectSpecFailure(spec.spec_uri, spec.triple_store["type"], None, message)
                    # if df.shape == then.shape and (df.columns == then.columns).all():
                    #     df_diff = then.compare(df, result_names=("expected", "actual"))
                    #     if df_diff.empty:
                    #         df_diff = df
                    #         print(df_diff.to_markdown())
                    # else:
                    #     df_diff = construct_df_diff(df, then)
                    #     print(df_diff.to_markdown())
                else:
                    if len(columns) == len(then.columns):
                        if sorted_columns == sorted_then_cols:
                            then = then[columns]
                            if not ordered_result:
                                then.sort_values(by=columns[::2], inplace=True)
                                then.reset_index(drop=True, inplace=True)
                            if df.shape == then.shape and (df.columns == then.columns).all():
                                df_diff = then.compare(df, result_names=("expected", "actual"))
                            else:
                                df_diff = construct_df_diff(df, then)

                        else:
                            then = then[sorted_then_cols]
                            df = df[sorted_columns]
                            df_diff = construct_df_diff(df, then)
                    else:

                        then = then[sorted_then_cols]
                        df = df[sorted_columns]
                        df_diff = construct_df_diff(df, then)
        else:

            if then.empty:
                # Scenario 3: expected no result, got no result
                message = f"Expected 0 row(s) and 0 column(s), got 0 row(s) and 0 column(s)"
                df = pandas.DataFrame()
            else:
                # Scenario 4: expected a result, but got an empty result
                message = f"Expected {then.shape[0]} row(s) and {round(then.shape[1] / 2)} column(s), got 0 row(s) and 0 column(s)"
                then = then[sorted_then_cols]
                df = create_empty_dataframe_with_columns(then)
            df_diff = then.compare(df, result_names=("expected", "actual"))
            print(df_diff.to_markdown())

        if df_diff.empty:
            if warning:
                return SpecPassedWithWarning(spec.spec_uri, spec.triple_store["type"], warning)
            else:
                return SpecPassed(spec.spec_uri, spec.triple_store["type"])
        else:
            # message += f"\nexpected:\n{then}\nactual:{df}"
            log.error(message)
            # print(spec.spec_uri)
            # print("actual:")
            # print(then)
            # print("expected:")
            # print(df)
            return SelectSpecFailure(spec.spec_uri, spec.triple_store["type"], df_diff, message)

    except ParseException as e:
        return SparqlParseFailure(spec.spec_uri, spec.triple_store["type"], e)
    except NotImplementedError as ex:
        return SpecSkipped(spec.spec_uri, spec.triple_store["type"], ex)


def graph_comparison(expected_graph: Graph, actual_graph: Graph) -> GraphComparison:
    diff = graph_diff(expected_graph, actual_graph)
    in_both = diff[0]
    in_expected = diff[1]
    in_actual = diff[2]
    in_expected_not_in_actual = (in_expected - in_actual)
    in_actual_not_in_expected = (in_actual - in_expected)
    return GraphComparison(in_expected_not_in_actual, in_actual_not_in_expected, in_both)


def get_then_update(spec_uri: URIRef, spec_graph: Graph) -> Graph:
    then_query = f"""
    prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> 

    CONSTRUCT {{ ?s ?p ?o }}
    {{
        <{spec_uri}> <{MUST.then}> 
            a <{MUST.StatementsDataset}> ;
            <{MUST.hasStatement}> [
                a rdf:Statement ;
                rdf:subject ?s ;
                rdf:predicate ?p ;
                rdf:object ?o ;
            ] ; ] 
    }}
    """
    expected_results = spec_graph.query(then_query).graph

    return expected_results


def calculate_row_difference(df1: pandas.DataFrame,
                             df2: pandas.DataFrame) -> pandas.DataFrame:
    df_all = df1.merge(df2.drop_duplicates(), how='left', indicator=True)
    actual_rows = df_all[df_all['_merge'] == 'left_only']
    actual_rows = actual_rows.drop('_merge', axis=1)
    return actual_rows


def construct_df_diff(df: pandas.DataFrame,
                      then: pandas.DataFrame) -> pandas.DataFrame:
    actual_rows = calculate_row_difference(df, then)
    expected_rows = calculate_row_difference(then, df)
    actual_columns = df.columns.difference(then.columns)
    expected_columns = then.columns.difference(df.columns)

    df_diff = pandas.DataFrame()
    modified_df = df
    modified_then = then

    if actual_columns.size > 0:
        modified_then = modified_then.reindex(modified_then.columns.to_list() + actual_columns.to_list(), axis=1)
        modified_then[actual_columns.to_list()] = modified_then[actual_columns.to_list()].fillna('')

    if expected_columns.size > 0:
        modified_df = modified_df.reindex(modified_df.columns.to_list() + expected_columns.to_list(), axis=1)
        modified_df[expected_columns.to_list()] = modified_df[expected_columns.to_list()].fillna('')

    modified_df = modified_df.reindex(modified_then.columns, axis=1)

    if df.shape[0] != then.shape[0] and df.shape[1] != then.shape[1]:
        # take modified columns and add rows
        actual_rows = calculate_row_difference(modified_df, modified_then)
        expected_rows = calculate_row_difference(modified_then, modified_df)
        df_diff = generate_row_diff(actual_rows, expected_rows)
    elif actual_rows.shape[0] > 0 or expected_rows.shape[0] > 0:
        df_diff = generate_row_diff(actual_rows, expected_rows)
    elif actual_columns.size > 0 or expected_columns.size > 0:
        df_diff = modified_then.compare(modified_df, result_names=("expected", "actual"), keep_shape=True,
                                        keep_equal=True)
    df_diff.fillna("", inplace=True)
    return df_diff


def generate_row_diff(actual_rows: pandas.DataFrame, expected_rows: pandas.DataFrame) -> pandas.DataFrame:
    df_diff_actual_rows = pandas.DataFrame()
    df_diff_expected_rows = pandas.DataFrame()

    if actual_rows.shape[0] > 0:
        empty_actual_copy = create_empty_dataframe_with_columns(actual_rows)
        df_diff_actual_rows = empty_actual_copy.compare(actual_rows, result_names=("expected", "actual"))

    if expected_rows.shape[0] > 0:
        empty_expected_copy = create_empty_dataframe_with_columns(expected_rows)
        df_diff_expected_rows = expected_rows.compare(empty_expected_copy, result_names=("expected", "actual"))

    df_diff_rows = pandas.concat([df_diff_actual_rows, df_diff_expected_rows], ignore_index=True)
    return df_diff_rows


def create_empty_dataframe_with_columns(df: pandas.DataFrame) -> pandas.DataFrame:
    empty_copy = pandas.DataFrame().reindex_like(df)
    empty_copy.fillna("", inplace=True)
    return empty_copy


def review_results(results: List[SpecResult], verbose: bool) -> None:
    print("===== Result Overview =====")
    # Init dictionaries
    status_dict = defaultdict(lambda: defaultdict(int))
    status_counts = defaultdict(lambda: defaultdict(int))
    colours = {SpecPassed: Fore.GREEN, SpecPassedWithWarning: Fore.YELLOW, SpecSkipped: Fore.YELLOW}
    # Populate dictionaries from results
    for result in results:
        status_counts[result.triple_store][type(result)] += 1
        status_dict[result.spec_uri][result.triple_store] = type(result)

    # Get the list of statuses and list of unique triple stores
    statuses = list(status for inner_dict in status_dict.values() for status in inner_dict.values())
    triple_stores = list(set(status for inner_dict in status_dict.values() for status in inner_dict.keys()))

    # Convert dictionaries to list for tabulate
    table_rows = [[spec_uri] + [
        f"{colours.get(status_dict[spec_uri][triple_store], Fore.RED)}{status_dict[spec_uri][triple_store].__name__}{Style.RESET_ALL}"
        for triple_store in triple_stores] for spec_uri in set(status_dict.keys())]

    status_rows = [[f"{colours.get(status, Fore.RED)}{status.__name__}{Style.RESET_ALL}"] +
                   [f"{colours.get(status, Fore.RED)}{status_counts[triple_store][status]}{Style.RESET_ALL}"
                    for triple_store in triple_stores] for status in set(statuses)]

    # Display tables with tabulate
    print(tabulate(table_rows, headers=['Spec Uris / triple stores'] + triple_stores, tablefmt="pretty"))
    print(tabulate(status_rows, headers=['Status / triple stores'] + triple_stores, tablefmt="pretty"))

    pass_count = statuses.count(SpecPassed)
    warning_count = statuses.count(SpecPassedWithWarning)
    skipped_count = statuses.count(SpecSkipped)
    fail_count = len(
        list(filter(lambda status: status not in [SpecPassed, SpecPassedWithWarning, SpecSkipped], statuses)))

    if fail_count:
        overview_colour = Fore.RED
    elif warning_count or skipped_count:
        overview_colour = Fore.YELLOW
    else:
        overview_colour = Fore.GREEN

    logger_setup.flush()
    print(f"{overview_colour}===== {fail_count} failures, {skipped_count} skipped, {Fore.GREEN}{pass_count} passed, "
          f"{overview_colour}{warning_count} passed with warnings =====")

    if verbose and (fail_count or warning_count or skipped_count):
        for res in results:
            if type(res) == UpdateSpecFailure:
                print(f"{Fore.RED}Failed {res.spec_uri} {res.triple_store}")
                print(f"{Fore.BLUE} In Expected Not In Actual:")
                print(res.graph_comparison.in_expected_not_in_actual.serialize(format="ttl"))
                print()
                print(f"{Fore.RED} in_actual_not_in_expected")
                print(res.graph_comparison.in_actual_not_in_expected.serialize(format="ttl"))
                print(f"{Fore.GREEN} in_both")
                print(res.graph_comparison.in_both.serialize(format="ttl"))

            if type(res) == SelectSpecFailure:
                print(f"{Fore.RED}Failed {res.spec_uri} {res.triple_store}")
                print(res.message)
                print(res.table_comparison.to_markdown())
            if type(res) == ConstructSpecFailure or type(res) == UpdateSpecFailure:
                print(f"{Fore.RED}Failed {res.spec_uri} {res.triple_store}")
            if type(res) == SpecPassedWithWarning:
                print(f"{Fore.YELLOW}Passed with warning {res.spec_uri} {res.triple_store}")
                print(res.warning)
            if type(res) == TripleStoreConnectionError or type(res) == SparqlExecutionError or \
                    type(res) == SparqlParseFailure:
                print(f"{Fore.RED}Failed {res.spec_uri} {res.triple_store}")
                print(res.exception)
            if type(res) == SpecSkipped:
                print(f"{Fore.YELLOW}Skipped {res.spec_uri} {res.triple_store}")
                print(res.message)



    





"""
MIT License

Copyright (c) 2023 Semantic Partners Ltd

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import os
from dataclasses import dataclass, field
from itertools import groupby
from pathlib import Path
from typing import Tuple, List, Type

import pandas
import requests
from rdflib import RDF, Graph, URIRef, Variable, Literal, XSD, util
from rdflib.exceptions import ParserError
from rdflib.term import Node
import logging 

import logger_setup
from mustrdAnzo import get_queries_for_layer, get_queries_from_templated_step, get_spec_component_from_graphmart, get_query_from_querybuilder, get_query_from_step
from namespace import MUST
from utils import get_project_root
from multimethods import MultiMethod, Default

log = logger_setup.setup_logger(__name__)


@dataclass
class SpecComponent:
    pass


@dataclass
class GivenSpec(SpecComponent):
    value: Graph = None


@dataclass
class WhenSpec(SpecComponent):
    value: str = None
    queryType: URIRef = None
    bindings: dict = None

@dataclass
class AnzoWhenSpec(WhenSpec):
    paramQuery: str = None
    queryTemplate: str = None

@dataclass
class ThenSpec(SpecComponent):
    value: Graph = Graph()
    ordered: bool = False


@dataclass
class TableThenSpec(ThenSpec):
    value: pandas.DataFrame = field(default_factory=pandas.DataFrame)


@dataclass
class SpecComponentDetails:
    subject: URIRef
    predicate: URIRef
    spec_graph: Graph
    mustrd_triple_store: dict
    spec_component_node: Node
    data_source_type: Node
    run_config: dict

def get_path(path_type: str, run_config: dict) -> Path:
    try:
        if str(run_config[path_type]).startswith("/"):
            return run_config[path_type]
        else: 
            return Path(os.path.join(run_config['spec_path'], run_config[path_type]))
    except(KeyError):
        if str(run_config['data_path']).startswith("/"):
            return run_config['data_path']
        else: 
            return Path(os.path.join(run_config['spec_path'], run_config['data_path']))
    

def parse_spec_component(subject: URIRef,
                         predicate: URIRef,
                         spec_graph: Graph,
                         run_config: dict,
                         mustrd_triple_store: dict) -> GivenSpec | WhenSpec | ThenSpec | TableThenSpec:
    # print(f"parse_spec_component {subject=} {predicate=} ")
    spec_component_nodes = get_spec_component_nodes(subject, predicate, spec_graph)
    # all_data_source_types = []
    spec_components = []
    for spec_component_node in spec_component_nodes:
        data_source_types = get_data_source_types(subject, predicate, spec_graph, spec_component_node)
        for data_source_type in data_source_types:
            spec_component_details = SpecComponentDetails(
                subject=subject,
                predicate=predicate,
                spec_graph=spec_graph,
                mustrd_triple_store=mustrd_triple_store,
                spec_component_node=spec_component_node,
                data_source_type=data_source_type,
                run_config=run_config)
            spec_component = get_spec_component(spec_component_details)
            if type(spec_component) == list:
                spec_components += spec_component 
            else:
                spec_components += [spec_component]

        # all_data_source_types.extend(data_source_types)
    # return all_data_source_types
    # merge multiple graphs into one, give error if spec config is a TableThen
    # print(f"calling multimethod with {spec_components}")
    return combine_specs(spec_components)


def get_spec_component_type(spec_components: List[SpecComponent]) -> Type[SpecComponent]:
    # Get the type of the first object in the list
    spec_type = type(spec_components[0])
    # Loop through the remaining objects in the list and check their types
    for spec_component in spec_components[1:]:
        if type(spec_component) != spec_type:
            # If an object has a different type, raise an error
            raise ValueError("All spec components must be of the same type")

    # If all objects have the same type, return the type
    return spec_type


def combine_specs_dispatch(spec_components: List[SpecComponent]) -> Type[SpecComponent]:
    spec_type = get_spec_component_type(spec_components)
    return spec_type


combine_specs = MultiMethod("combine_specs", combine_specs_dispatch)


@combine_specs.method(GivenSpec)
def _combine_given_specs(spec_components: List[GivenSpec]) -> GivenSpec:
    if len(spec_components) == 1:
        return spec_components[0]
    else:
        graph = Graph()
        for spec_component in spec_components:
            graph += spec_component.value
        given_spec = GivenSpec()
        given_spec.value = graph
        return given_spec


@combine_specs.method(WhenSpec)
def _combine_when_specs(spec_components: List[WhenSpec]) -> WhenSpec:
    return spec_components


@combine_specs.method(ThenSpec)
def _combine_then_specs(spec_components: List[ThenSpec]) -> ThenSpec:
    if len(spec_components) == 1:
        return spec_components[0]
    else:
        graph = Graph()
        for spec_component in spec_components:
            graph += spec_component.value
        then_spec = ThenSpec()
        then_spec.value = graph
        return then_spec


@combine_specs.method(TableThenSpec)
def _combine_table_then_specs(spec_components: List[TableThenSpec]) -> TableThenSpec:
    if len(spec_components) != 1:
        raise ValueError(f"Parsing of multiple components of MUST.then for tables not implemented")
    return spec_components[0]


@combine_specs.method(Default)
def _combine_specs_default(spec_components: List[SpecComponent]):
    raise ValueError(f"Parsing of multiple components of this type not implemented {spec_components}")

def get_data_source_types(subject: URIRef, predicate: URIRef, spec_graph: Graph, source_node: Node) -> List[Node]:
    data_source_types = []
    for data_source_type in spec_graph.objects(subject=source_node, predicate=RDF.type):
        data_source_types.append(data_source_type)
    # data_source_type = spec_graph.value(subject=source_node, predicate=RDF.type)
    if len(data_source_types) == 0:
        raise ValueError(f"Node has no rdf type {subject} {predicate}")
    return data_source_types

# https://github.com/Semantic-partners/mustrd/issues/99
def get_spec_component_dispatch(spec_component_details: SpecComponentDetails) -> Tuple[Node, URIRef]:
    return spec_component_details.data_source_type, spec_component_details.predicate


get_spec_component = MultiMethod("get_spec_component", get_spec_component_dispatch)


@get_spec_component.method((MUST.InheritedDataset, MUST.given))
def _get_spec_component_inheritedstate_given(spec_component_details: SpecComponentDetails) -> GivenSpec:
    spec_component = init_spec_component(spec_component_details.predicate)
    return spec_component


@get_spec_component.method((MUST.FolderDataset, MUST.given))
def _get_spec_component_folderdatasource_given(spec_component_details: SpecComponentDetails) -> GivenSpec:
    spec_component = init_spec_component(spec_component_details.predicate)

    file_name = spec_component_details.spec_graph.value(subject=spec_component_details.spec_component_node,
                                                        predicate=MUST.fileName)

    path = Path(os.path.join(str(get_path('given_path',spec_component_details.run_config)), str(file_name)))
    try:
        spec_component.value = Graph().parse(data=get_spec_component_from_file(path))
    except ParserError as e:
        log.error(f"Problem parsing {path}, error of type {type(e)}")
        raise ValueError(f"Problem parsing {path}, error of type {type(e)}")
    return spec_component


@get_spec_component.method((MUST.FolderSparqlSource, MUST.when))
def _get_spec_component_foldersparqlsource_when(spec_component_details: SpecComponentDetails) -> GivenSpec:
    spec_component = init_spec_component(spec_component_details.predicate)

    file_name = spec_component_details.spec_graph.value(subject=spec_component_details.spec_component_node,
                                                        predicate=MUST.fileName)

    path = Path(os.path.join(str(get_path('when_path',spec_component_details.run_config)), str(file_name)))
    spec_component.value = get_spec_component_from_file(path)
    spec_component.queryType = spec_component_details.spec_graph.value(subject=spec_component_details.spec_component_node,
                                                                       predicate=MUST.queryType)
    return spec_component


@get_spec_component.method((MUST.FolderDataset, MUST.then))
def _get_spec_component_folderdatasource_then(spec_component_details: SpecComponentDetails) -> ThenSpec:
    spec_component = init_spec_component(spec_component_details.predicate)

    file_name = spec_component_details.spec_graph.value(subject=spec_component_details.spec_component_node,
                                                        predicate=MUST.fileName)
    path = Path(os.path.join(str(get_path('then_path',spec_component_details.run_config)), str(file_name)))

    return load_dataset_from_file(path, spec_component)

@get_spec_component.method((MUST.FileDataset, MUST.given))
@get_spec_component.method((MUST.FileDataset, MUST.then))
def _get_spec_component_filedatasource(spec_component_details: SpecComponentDetails) -> GivenSpec:
    spec_component = init_spec_component(spec_component_details.predicate)
    return load_spec_component(spec_component_details, spec_component)

def load_spec_component(spec_component_details, spec_component):
    where_did_i_load_this_spec_from = spec_component_details.spec_graph.value(subject=spec_component_details.subject,
                                                                 predicate=MUST.specSourceFile)
    if (where_did_i_load_this_spec_from == None):
        log.error(f"{where_did_i_load_this_spec_from=} was None for test_spec={spec_component_details.subject}, we didn't set the test specifications specSourceFile when loading, spec_graph={spec_component_details.spec_graph}")
    file_path = Path(str(spec_component_details.spec_graph.value(subject=spec_component_details.spec_component_node,
                                                                 predicate=MUST.file)))
    
    test_spec_file_path = os.path.dirname(where_did_i_load_this_spec_from)

    # first we try local relative to the test_spec_file_path, then we try relative to the path under test
    # we intentionally don't try for absolute files, but you should feel free to argue that we should do.
    paths = [
        Path(test_spec_file_path, file_path),
        Path(os.path.join(spec_component_details.run_config['spec_path'], file_path))
    ]
    
    for path in paths:
        if (os.path.exists(path)):
            return load_dataset_from_file(path, spec_component)

    raise FileNotFoundError(f"Could not find file {file_path=} in any of the {paths=}")
    

def load_dataset_from_file(path: Path, spec_component: ThenSpec) -> ThenSpec:
    if path.is_dir():
        raise ValueError(f"Path {path} is a directory, expected a file")

    # https://github.com/Semantic-partners/mustrd/issues/94
    if path.suffix in {".csv", ".xlsx", ".xls"}:
        df = pandas.read_csv(path) if path.suffix == ".csv" else pandas.read_excel(path)
        then_spec = TableThenSpec()
        then_spec.value = df
        return then_spec
    else:
        try:
            file_format = util.guess_format(str(path))
        except AttributeError:
            raise ValueError(f"Unsupported file format: {path.suffix}")

        if file_format is not None:
            g = Graph()
            try:
                g.parse(data=get_spec_component_from_file(path), format=file_format)
            except ParserError as e:
                log.error(f"Problem parsing {path}, error of type {type(e)}")
                raise ValueError(f"Problem parsing {path}, error of type {type(e)}")
            spec_component.value = g
            return spec_component



@get_spec_component.method((MUST.FileSparqlSource, MUST.when))
def _get_spec_component_filedatasource_when(spec_component_details: SpecComponentDetails) -> SpecComponent:
    spec_component = init_spec_component(spec_component_details.predicate)

    file_path = Path(str(spec_component_details.spec_graph.value(subject=spec_component_details.spec_component_node,
                                                                 predicate=MUST.file)))
    if str(file_path).startswith("/"): # absolute path
        path = file_path
    else: #relative path
        path = Path(os.path.join(spec_component_details.run_config['spec_path'], file_path))
    spec_component.value = get_spec_component_from_file(path)

    spec_component.queryType = spec_component_details.spec_graph.value(subject=spec_component_details.spec_component_node,
                                                                       predicate=MUST.queryType)

    return spec_component


# @get_spec_component.method((MUST.FileDataset, MUST.then))
# def _get_spec_component_filedatasource_then(spec_component_details: SpecComponentDetails) -> SpecComponent:
#     spec_component = init_spec_component(spec_component_details.predicate)

#     file_path = Path(str(spec_component_details.spec_graph.value(subject=spec_component_details.spec_component_node,
#                                                                  predicate=MUST.file)))
#     if str(file_path).startswith("/"): # absolute path
#         path = file_path
#     else: #relative path
#         path = Path(os.path.join(spec_component_details.run_config['spec_path'], file_path))
#     return get_then_from_file(path, spec_component)


@get_spec_component.method((MUST.TextSparqlSource, MUST.when))
def _get_spec_component_TextSparqlSource(spec_component_details: SpecComponentDetails) -> SpecComponent:
    spec_component = init_spec_component(spec_component_details.predicate)

    # Get specComponent directly from config file (in text string)
    spec_component.value = str(
        spec_component_details.spec_graph.value(subject=spec_component_details.spec_component_node,
                                                predicate=MUST.queryText))

    spec_component.bindings = get_when_bindings(spec_component_details.subject, spec_component_details.spec_graph)
    spec_component.queryType = spec_component_details.spec_graph.value(subject=spec_component_details.spec_component_node,
                                                                       predicate=MUST.queryType)
    return spec_component


# https://github.com/Semantic-partners/mustrd/issues/98
@get_spec_component.method((MUST.HttpDataset, MUST.given))
@get_spec_component.method((MUST.HttpDataset, MUST.when))
@get_spec_component.method((MUST.HttpDataset, MUST.then))
def _get_spec_component_HttpDataset(spec_component_details: SpecComponentDetails) -> SpecComponent:
    spec_component = init_spec_component(spec_component_details.predicate)

    # Get specComponent with http GET protocol
    spec_component.value = requests.get(str(
        spec_component_details.spec_graph.value(subject=spec_component_details.spec_component_node,
                                                predicate=MUST.dataSourceUrl)).content)    
    spec_component.queryType = spec_component_details.spec_graph.value(subject=spec_component_details.spec_component_node,
                                                                       predicate=MUST.queryType)
    return spec_component


@get_spec_component.method((MUST.TableDataset, MUST.then))
def _get_spec_component_TableDataset(spec_component_details: SpecComponentDetails) -> SpecComponent:
    table_then = TableThenSpec()
    # get specComponent from ttl table
    table_then.value = get_spec_from_table(spec_component_details.subject, spec_component_details.predicate,
                                           spec_component_details.spec_graph)
    table_then.ordered = is_then_select_ordered(spec_component_details.subject, spec_component_details.predicate,
                                                spec_component_details.spec_graph)
    return table_then


@get_spec_component.method((MUST.EmptyTable, MUST.then))
def _get_spec_component_EmptyTable(spec_component_details: SpecComponentDetails) -> SpecComponent:
    spec_component = TableThenSpec()
    return spec_component


@get_spec_component.method((MUST.EmptyGraph, MUST.then))
def _get_spec_component_EmptyGraph(spec_component_details: SpecComponentDetails) -> SpecComponent:
    spec_component = init_spec_component(spec_component_details.predicate)

    return spec_component


@get_spec_component.method((MUST.StatementsDataset, MUST.given))
@get_spec_component.method((MUST.StatementsDataset, MUST.then))
def _get_spec_component_StatementsDataset(spec_component_details: SpecComponentDetails) -> SpecComponent:
    spec_component = init_spec_component(spec_component_details.predicate)

    spec_component.value = Graph().parse(
        data=get_spec_from_statements(spec_component_details.subject, spec_component_details.predicate,
                                      spec_component_details.spec_graph))
    return spec_component


@get_spec_component.method((MUST.AnzoGraphmartDataset, MUST.given))
@get_spec_component.method((MUST.AnzoGraphmartDataset, MUST.then))
def _get_spec_component_AnzoGraphmartDataset(spec_component_details: SpecComponentDetails) -> SpecComponent:
    spec_component = init_spec_component(spec_component_details.predicate)

    if spec_component_details.mustrd_triple_store["type"] == MUST.Anzo:
        # Get GIVEN or THEN from anzo graphmart
        graphmart = spec_component_details.spec_graph.value(subject=spec_component_details.spec_component_node,
                                                            predicate=MUST.graphmart)
        layer = spec_component_details.spec_graph.value(subject=spec_component_details.spec_component_node,
                                                        predicate=MUST.layer)
        spec_component.value = get_spec_component_from_graphmart(
            triple_store=spec_component_details.mustrd_triple_store,
            graphmart=graphmart,
            layer=layer)
    else:
        raise ValueError(f"You must define {MUST.AnzoConfig} to use {MUST.AnzoGraphmartDataset}")

    return spec_component


@get_spec_component.method((MUST.AnzoQueryBuilderSparqlSource, MUST.when))
def _get_spec_component_AnzoQueryBuilderSparqlSource(spec_component_details: SpecComponentDetails) -> SpecComponent:
    spec_component = init_spec_component(spec_component_details.predicate)

    # Get WHEN specComponent from query builder
    if spec_component_details.mustrd_triple_store["type"] == MUST.Anzo:
        query_folder = spec_component_details.spec_graph.value(subject=spec_component_details.spec_component_node,
                                                               predicate=MUST.queryFolder)
        query_name = spec_component_details.spec_graph.value(subject=spec_component_details.spec_component_node,
                                                             predicate=MUST.queryName)
        spec_component.value = get_query_from_querybuilder(triple_store=spec_component_details.mustrd_triple_store,
                                                           folder_name=query_folder,
                                                           query_name=query_name)
    # If anzo specific function is called but no anzo defined
    else:
        raise ValueError(f"You must define {MUST.AnzoConfig} to use {MUST.AnzoQueryBuilderSparqlSource}")

    spec_component.queryType = spec_component_details.spec_graph.value(subject=spec_component_details.spec_component_node,
                                                                       predicate=MUST.queryType)
    return spec_component


@get_spec_component.method((MUST.AnzoGraphmartStepSparqlSource, MUST.when))
def _get_spec_component_AnzoGraphmartStepSparqlSource(spec_component_details: SpecComponentDetails) -> SpecComponent:
    spec_component = init_spec_component(spec_component_details.predicate)

    # Get WHEN specComponent from query builder
    if spec_component_details.mustrd_triple_store["type"] == MUST.Anzo:
        query_step_uri = spec_component_details.spec_graph.value(subject=spec_component_details.spec_component_node,
                                                             predicate=MUST.anzoQueryStep)
        spec_component.value = get_query_from_step(triple_store=spec_component_details.mustrd_triple_store,
                                                    query_step_uri=query_step_uri)
    # If anzo specific function is called but no anzo defined
    else:
        raise ValueError(f"You must define {MUST.AnzoConfig} to use {MUST.AnzoGraphmartStepSparqlSource}")

    spec_component.queryType = spec_component_details.spec_graph.value(subject=spec_component_details.spec_component_node,
                                                                       predicate=MUST.queryType)
    return spec_component

@get_spec_component.method((MUST.AnzoGraphmartQueryDrivenTemplatedStepSparqlSource, MUST.when))
def _get_spec_component_AnzoGraphmartQueryDrivenTemplatedStepSparqlSource(spec_component_details: SpecComponentDetails) -> SpecComponent:
    spec_component = init_spec_component(spec_component_details.predicate, spec_component_details.mustrd_triple_store["type"] )

    # Get WHEN specComponent from query builder
    if spec_component_details.mustrd_triple_store["type"] == MUST.Anzo:
        query_step_uri = spec_component_details.spec_graph.value(subject=spec_component_details.spec_component_node,
                                                             predicate=MUST.anzoQueryStep)
        queries = get_queries_from_templated_step(triple_store=spec_component_details.mustrd_triple_store,
                                                    query_step_uri=query_step_uri)
        spec_component.paramQuery= queries["param_query"]
        spec_component.queryTemplate = queries["query_template"]
    # If anzo specific function is called but no anzo defined
    else:
        raise ValueError(f"You must define {MUST.AnzoConfig} to use {MUST.AnzoGraphmartQueryDrivenTemplatedStepSparqlSource}")

    spec_component.queryType = spec_component_details.spec_graph.value(subject=spec_component_details.spec_component_node,
                                                                       predicate=MUST.queryType)
    return spec_component

@get_spec_component.method((MUST.AnzoGraphmartLayerSparqlSource, MUST.when))
def _get_spec_component_AnzoGraphmartLayerSparqlSource(spec_component_details: SpecComponentDetails) -> list:
    spec_components = []
    # Get the ordered  WHEN specComponents which is the transform and query driven template queries for the Layer
    if spec_component_details.mustrd_triple_store["type"] == MUST.Anzo:
        graphmart_layer_uri = spec_component_details.spec_graph.value(subject=spec_component_details.spec_component_node,
                                                             predicate=MUST.anzoGraphmartLayer)
        queries = get_queries_for_layer(triple_store=spec_component_details.mustrd_triple_store,
                                                    graphmart_layer_uri=graphmart_layer_uri)
    # If anzo specific function is called but no anzo defined
    else:
        raise ValueError(f"This test specification is specific to Anzo and can only be run against that platform.")
    for query in queries:
        spec_component = init_spec_component(spec_component_details.predicate, spec_component_details.mustrd_triple_store["type"])
        spec_component.value = query.get("query") 
        spec_component.paramQuery = query.get("param_query")
        spec_component.queryTemplate = query.get("query_template")
        if spec_component.value:
            spec_component.queryType = spec_component_details.spec_graph.value(subject=spec_component_details.spec_component_node,
                                                                       predicate=MUST.queryType)
        else:
            spec_component.queryType  = MUST.AnzoQueryDrivenUpdateSparql   
        spec_components += [spec_component]
    return spec_components

@get_spec_component.method(Default)
def _get_spec_component_default(spec_component_details: SpecComponentDetails) -> SpecComponent:
    raise ValueError(
        f"Invalid combination of data source type ({spec_component_details.data_source_type}) and "
        f"spec component ({spec_component_details.predicate})")


def init_spec_component(predicate: URIRef, triple_store_type: URIRef = None ) -> GivenSpec | WhenSpec | ThenSpec | TableThenSpec:
    if predicate == MUST.given:
        spec_component = GivenSpec()
    elif predicate == MUST.when:
        if triple_store_type == MUST.Anzo:
            spec_component = AnzoWhenSpec()
        else:
            spec_component = WhenSpec()
    elif predicate == MUST.then:
        spec_component = ThenSpec()
    else:
        spec_component = SpecComponent()
    return spec_component


def get_spec_component_nodes(subject: URIRef, predicate: URIRef, spec_graph: Graph) -> List[Node]:
    spec_component_nodes = []
    for spec_component_node in spec_graph.objects(subject=subject, predicate=predicate):
        spec_component_nodes.append(spec_component_node)
    # It shouldn't even be possible to get this far as an empty node indicates an invalid RDF file
    if spec_component_nodes is None:
        raise ValueError(f"specComponent Node empty for {subject} {predicate}")
    return spec_component_nodes


def get_spec_component_from_file(path: Path) -> str:
    # project_root = get_project_root()
    # file_path = Path(os.path.join(project_root, path))

    if path.is_dir():
        raise ValueError(f"Path {path} is a directory, expected a file")

    try:
        content = path.read_text()
    except FileNotFoundError:
        raise
    return str(content)


def get_spec_from_statements(subject: URIRef,
                             predicate: URIRef,
                             spec_graph: Graph) -> Graph:
    statements_query = f"""
    prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> 

    CONSTRUCT {{ ?s ?p ?o }}
    {{
            <{subject}> <{predicate}> [
                a <{MUST.StatementsDataset}> ;
                <{MUST.hasStatement}> [
                    a rdf:Statement ;
                    rdf:subject ?s ;
                    rdf:predicate ?p ;
                    rdf:object ?o ;
                ] ;
            ]

    }}
    """
    results = spec_graph.query(statements_query).graph
    return results.serialize(format="ttl")


def get_spec_from_table(subject: URIRef,
                        predicate: URIRef,
                        spec_graph: Graph) -> pandas.DataFrame:
    # query the spec to get the expected result to convert to dataframe for comparison
    then_query = f"""
        prefix sh:        <http://www.w3.org/ns/shacl#> 
            SELECT ?row ?variable ?binding ?order
            WHERE {{ 
                 <{subject}> <{predicate}> [
                        a <{MUST.TableDataset}> ;
                        <{MUST.hasRow}> ?row ].
                          ?row  <{MUST.hasBinding}> [
                                <{MUST.variable}> ?variable ;
                                <{MUST.boundValue}> ?binding ; ] .
                          OPTIONAL {{ ?row sh:order ?order . }}        
                                     .}} 
             ORDER BY ?order"""

    expected_results = spec_graph.query(then_query)
    # get the unique row ids form the result to form the index of the results dataframe
    index = {str(row.row) for row in expected_results}
    # get the unique variables to form the columns of the results dataframe
    columns = set()
    for row in expected_results:
        columns.add(row.variable.value)
        columns.add(row.variable.value + "_datatype")
    # add an additional column for the sort order (if any) of the results
    columns.add("order")
    # create an empty dataframe to populate with the results
    df = pandas.DataFrame(index=list(index), columns=list(columns))
    # fill the dataframe with the results data
    for row in expected_results:
        df.loc[str(row.row), row.variable.value] = str(row.binding)
        df.loc[str(row.row), "order"] = row.order
        if type(row.binding) == Literal:
            literal_type = str(XSD.string)
            if hasattr(row.binding, "datatype") and row.binding.datatype:
                literal_type = str(row.binding.datatype)
            df.loc[str(row.row), row.variable.value + "_datatype"] = literal_type
        else:
            df.loc[str(row.row), row.variable.value + "_datatype"] = str(XSD.anyURI)
    # use the sort order sort the results
    df.sort_values(by="order", inplace=True)
    # drop the order column and replace the rowid index with a numeric one and replace empty values with spaces
    df.drop(columns="order", inplace=True)
    df.reset_index(drop=True, inplace=True)
    df.fillna('', inplace=True)
    return df


def get_when_bindings(subject: URIRef,
                      spec_graph: Graph) -> dict:
    when_bindings_query = f"""SELECT ?variable ?binding {{ <{subject}> <{MUST.when}> [ a <{MUST.TextSparqlSource}> ; <{MUST.hasBinding}> [ <{MUST.variable}> ?variable ; <{MUST.boundValue}> ?binding ; ] ; ]  ;}}"""
    when_bindings = spec_graph.query(when_bindings_query)

    if len(when_bindings.bindings) == 0:
        return {}
    else:
        bindings = {}
        for binding in when_bindings:
            bindings[Variable(binding.variable.value)] = binding.binding
        return bindings


def is_then_select_ordered(subject: URIRef, predicate: URIRef, spec_graph: Graph) -> bool:
    ask_select_ordered = f"""
    ASK {{
    {{SELECT (count(?binding) as ?totalBindings) {{  
    <{subject}> <{predicate}> [
                a <{MUST.TableDataset}> ;
                <{MUST.hasRow}> [ <{MUST.hasBinding}> [
                                    <{MUST.variable}> ?variable ;
                                    <{MUST.boundValue}> ?binding ;
                            ] ; 
              ]
            ]
}} }}
    {{SELECT (count(?binding) as ?orderedBindings) {{    
    <{subject}> <{predicate}> [
                a <{MUST.TableDataset}> ;
       <{MUST.hasRow}> [ sh:order ?order ;
                    <{MUST.hasBinding}> [ 
                    <{MUST.variable}> ?variable ;
                                    <{MUST.boundValue}> ?binding ;
                            ] ; 
              ]
            ]
}} }}
    FILTER(?totalBindings = ?orderedBindings)
}}"""
    is_ordered = spec_graph.query(ask_select_ordered)
    return is_ordered.askAnswer



"""
MIT License

Copyright (c) 2023 Semantic Partners Ltd

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from dataclasses import dataclass
from TestResult import ResultList, TestResult, get_result_list
import pytest
import os
from pathlib import Path
from rdflib.namespace import Namespace
from rdflib import Graph

from utils import get_project_root
from mustrd import get_triple_store_graph, get_triple_stores, SpecSkipped, validate_specs, get_specs, SpecPassed, run_spec
from namespace import MUST
from pytest import Session
from typing import Dict

spnamespace = Namespace("https://semanticpartners.com/data/test/")

project_root = get_project_root()


@dataclass
class TestConfig:
    test_function: str
    spec_path: str
    data_path: str
    triplestore_spec_path: str
    filter_on_tripleStore: str

    def __init__(self, test_function: str, spec_path: str, data_path: str, triplestore_spec_path: str,
                 filter_on_tripleStore: str = None):
        self.test_function = test_function
        self.spec_path = spec_path
        self.data_path = data_path
        self.triplestore_spec_path = triplestore_spec_path
        self.filter_on_tripleStore = filter_on_tripleStore


class MustrdTestPlugin:
    md_path: str
    test_configs: Dict[str, TestConfig]

    def __init__(self, md_path, test_configs):
        self.md_path = md_path
        self.test_configs = test_configs

    # Hook called at collection time: reads the configuration of the tests, and generate pytests from it
    def pytest_generate_tests(self, metafunc):

        if len(metafunc.fixturenames) > 0:
            if metafunc.function.__name__ in self.test_configs:
                one_test_config = self.test_configs[metafunc.function.__name__]

                triple_stores = self.get_triple_stores_from_file(one_test_config)

                unit_tests = []
                if one_test_config.filter_on_tripleStore and not triple_stores:
                    unit_tests = list(map(lambda triple_store:
                                      SpecSkipped(MUST.TestSpec, triple_store, "No triplestore found"),
                                      one_test_config.filter_on_tripleStore))
                else:
                    unit_tests = self.generate_tests_for_config({"spec_path": project_root / one_test_config.spec_path,
                                                                "data_path": project_root / one_test_config.data_path},
                                                                triple_stores)

                # Create the test in itself
                if unit_tests:
                    metafunc.parametrize(metafunc.fixturenames[0], unit_tests, ids=self.get_test_name)
            else:
                metafunc.parametrize(metafunc.fixturenames[0],
                                     [SpecSkipped(MUST.TestSpec, None, "No triplestore found")],
                                     ids=lambda x: "No configuration found for this test")

    # Generate test for each triple store available
    def generate_tests_for_config(self, config, triple_stores):

        shacl_graph = Graph().parse(Path(os.path.join(project_root, "model/mustrdShapes.ttl")))
        ont_graph = Graph().parse(Path(os.path.join(project_root, "model/ontology.ttl")))
        valid_spec_uris, spec_graph, invalid_spec_results = validate_specs(config, triple_stores,
                                                                           shacl_graph, ont_graph)

        specs, skipped_spec_results = \
            get_specs(valid_spec_uris, spec_graph, triple_stores, config)

        # Return normal specs + skipped results
        return specs + skipped_spec_results + invalid_spec_results

    # Function called to generate the name of the test
    def get_test_name(self, spec):
        # FIXME: SpecSkipped should have the same structure?
        if isinstance(spec, SpecSkipped):
            triple_store = spec.triple_store
        else:
            triple_store = spec.triple_store['type']
        triple_store_name = triple_store.replace("https://mustrd.com/model/", "")
        test_name = spec.spec_uri.replace(spnamespace, "").replace("_", " ")
        return triple_store_name + ": " + test_name

    # Get triple store configuration or default
    def get_triple_stores_from_file(self, test_config):
        if test_config.triplestore_spec_path:
            try:
                triple_stores = get_triple_stores(get_triple_store_graph(project_root / test_config.triplestore_spec_path))
            except Exception:
                print(f"""No triple store configuration found at {project_root / test_config.triplestore_spec_path}.
                    Fall back: only embedded rdflib will be executed""")
                triple_stores = [{'type': MUST.RdfLib}]
        else:
            print("No triple store configuration required: using embedded rdflib")
            triple_stores = [{'type': MUST.RdfLib}]

        if test_config.filter_on_tripleStore:
            triple_stores = list(filter(lambda triple_store: (triple_store["type"] in test_config.filter_on_tripleStore),
                                        triple_stores))
        return triple_stores

    # Hook function. Initialize the list of result in session
    def pytest_sessionstart(self, session):
        session.results = dict()

    # Hook function called each time a report is generated by a test
    # The report is added to a list in the session
    # so it can be used later in pytest_sessionfinish to generate the global report md file
    @pytest.hookimpl(tryfirst=True, hookwrapper=True)
    def pytest_runtest_makereport(self, item, call):
        outcome = yield
        result = outcome.get_result()

        if result.when == 'call':
            # Add the result of the test to the session
            item.session.results[item] = result

    # Take all the test results in session, parse them, split them in mustrd and standard pytest  and generate md file
    def pytest_sessionfinish(self, session: Session, exitstatus):
        # if md path has not been defined in argument, then do not generate md file
        if not self.md_path:
            return

        test_results = []
        for test_conf, result in session.results.items():
            # Case auto generated tests
            if test_conf.originalname != test_conf.name:
                module_name = test_conf.parent.name
                class_name = test_conf.originalname
                test_name = test_conf.name.replace(class_name, "").replace("[", "").replace("]", "")
                is_mustrd = True
            # Case normal unit tests
            else:
                module_name = test_conf.parent.parent.name
                class_name = test_conf.parent.name
                test_name = test_conf.originalname
                is_mustrd = False

            test_results.append(TestResult(test_name, class_name, module_name, result.outcome, is_mustrd))

        result_list = ResultList(None, get_result_list(test_results,
                                                       lambda result: result.type,
                                                       lambda result: result.module_name,
                                                       lambda result: result.class_name),
                                 False)

        md = result_list.render()
        with open(self.md_path, 'w') as file:
            file.write(md)


# Function called in the test to actually run it
def run_test_spec(test_spec):
    if isinstance(test_spec, SpecSkipped):
        pytest.skip(f"Invalid configuration, error : {test_spec.message}")
    result = run_spec(test_spec)

    result_type = type(result)
    if result_type == SpecSkipped:
        # FIXME: Better exception management
        pytest.skip("Unsupported configuration")
    return result_type == SpecPassed



"""
MIT License

Copyright (c) 2023 Semantic Partners Ltd

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import argparse
import logger_setup
import sys
import os
from rdflib import Graph
from mustrd import get_triple_store_graph, run_specs, get_triple_stores, review_results, validate_specs, get_specs
from pathlib import Path
from namespace import MUST
from utils import get_project_root
import logging 
log = logger_setup.setup_logger(__name__)


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument("-p", "--put", help="Path under test - required", required=True)
    parser.add_argument("-v", "--verbose", help="verbose logging", action='store_true')
    parser.add_argument("-c", "--config", help="Path to triple store configuration", default=None)
    parser.add_argument("-d", "--data", help="Path to test data", default=None)
    parser.add_argument("-g", "--given", help="Override path for given files", default=None)
    parser.add_argument("-w", "--when", help="Override path for when files", default=None)
    parser.add_argument("-t", "--then", help="Override path for then files", default=None)

    return parser.parse_args()


# https://github.com/Semantic-partners/mustrd/issues/108
def main(argv):
    #given_path = when_path = then_path = None
    project_root = get_project_root()
    run_config = {}
    args = parse_args()
    run_config["spec_path"] = Path(args.put)
    log.info(f"Path under test is { run_config['spec_path']}")

    verbose = args.verbose
    if verbose:
        log.info(f"Verbose set")
        run_config["verbose"] = True

    if args.config:
        triplestore_spec_path = Path(args.config)
        log.info(f"Path for triple store configuration is {triplestore_spec_path}")
        triple_stores = get_triple_stores(get_triple_store_graph(triplestore_spec_path))
    else:
        log.info(f"No triple store configuration added, running default configuration")
        triple_stores = [{'type': MUST.RdfLib}]
    log.info("Triple Stores: " + str(triple_stores))
    if args.data:
        run_config["data_path"] = Path(args.data)
    else:
        run_config["data_path"] = Path(args.put)
    log.info(f"Path for test data folder is {run_config['data_path']}")

    if args.given:
        run_config["given_path"] = Path(args.given)
        log.info(f"Path for given folder is {run_config['given_path']}")

    if args.when:
        run_config["when_path"] = Path(args.when)
        log.info(f"Path for when folder is {run_config['when_path']}")

    if args.then:
        run_config["then_path"] = Path(args.then)
        log.info(f"Path for then folder is {run_config['then_path']}")

    shacl_graph = Graph().parse(Path(os.path.join(project_root, "model/mustrdShapes.ttl")))
    ont_graph = Graph().parse(Path(os.path.join(project_root, "model/ontology.ttl")))

    valid_spec_uris, spec_graph, invalid_spec_results = \
        validate_specs(run_config, triple_stores, shacl_graph, ont_graph)

    specs, skipped_spec_results = \
        get_specs(valid_spec_uris, spec_graph, triple_stores, run_config)

    results = invalid_spec_results + skipped_spec_results + run_specs(specs)

    review_results(results, verbose)


if __name__ == "__main__":
    main(sys.argv[1:])
